#include "CHMpreCompHead.h"
#pragma hdrstop

#ifdef CHM_NO_PRECOMPILED_HEADERS
#   include <CHM/CHMdateTimeClass.h>
#   include <CHM/CHMdateTime.h>
#   include <CHM/CHMerrorClass.h>
#   include <math.h>
#endif

#define THROW_DATE_TIME_ERROR throw CHMerror(CHM_TEXT("Date time error."), __LINE__, __FILE__, 0);

class CHMdateTimePrivate
{
public:
   CHMdateTimePrivate();
   CHMdateTimePrivate(CHMdateTimeHandle Handle);
   ~CHMdateTimePrivate();
   CHMboolean SetDATEFromTm(int wYear, int wMonth, int wDay, 
                            int wHour, int wMinute, int wSecond, 
                            DATE& dtDest);

   CHMboolean SetTmFromDATE(DATE dtSrc, struct tm& tmDest);

   CHMdateTimeHandle Handle; 
};

CHMdateTimePrivate::CHMdateTimePrivate()
 : Handle(NULL)
{
   {
      CHM_CHECK_CALL(CHMdateTimeCreate(&Handle));
   }
   {
      CHM_CHECK_CALL(CHMdateTimeAddRef(Handle));
   }
}

CHMdateTimePrivate::CHMdateTimePrivate(CHMdateTimeHandle iHandle)
 : Handle(iHandle) 
{
   CHM_CHECK_CALL(CHMdateTimeAddRef(Handle));
}

CHMdateTimePrivate::~CHMdateTimePrivate()
{
   CHM_CHECK_CALL(CHMdateTimeRelease(Handle));
}

// Verifies will fail if the needed buffer size is too large
#define MAX_TIME_BUFFER_SIZE    128         // matches that in timecore.cpp
#define MIN_DATE                (-657434L)  // about year 100
#define MAX_DATE                2958465L    // about year 9999

// Half a second, expressed in days
#define HALF_SECOND  (1.0/172800.0)

// One-based array of days in year at month start
int CHMMonthDays[13] =
   {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365};

CHMboolean CHMdateTimePrivate::SetDATEFromTm
(
   int wYear, 
   int wMonth, 
   int wDay,
   int wHour, 
   int wMinute, 
   int wSecond, 
   DATE& dtDest
)
{
   // Validate year and month (ignore day of week and milliseconds)
   if (wYear > 9999 || wMonth < 1 || wMonth > 12)
      return false;

   //  Check for leap year and set the number of days in the month
   CHMboolean bLeapYear = ((wYear & 3) == 0) &&
      ((wYear % 100) != 0 || (wYear % 400) == 0);

   int nDaysInMonth =
      CHMMonthDays[wMonth] - CHMMonthDays[wMonth-1] +
      ((bLeapYear && wDay == 29 && wMonth == 2) ? 1 : 0);

   // Finish validating the date
   if (wDay < 1 || wDay > nDaysInMonth ||
      wHour > 23 || wMinute > 59 ||
      wSecond > 59)
   {
      return false;
   }

   // Cache the date in days and time in fractional days
   long nDate;
   double dblTime;

   //It is a valid date; make Jan 1, 1AD be 1
   nDate = wYear*365L + wYear/4 - wYear/100 + wYear/400 +
      CHMMonthDays[wMonth-1] + wDay;

   //  If leap year and it's before March, subtract 1:
   if (wMonth <= 2 && bLeapYear)
      --nDate;

   //  Offset so that 12/30/1899 is 0
   nDate -= 693959L;

   dblTime = (((long)wHour * 3600L) +  // hrs in seconds
      ((long)wMinute * 60L) +  // mins in seconds
      ((long)wSecond)) / 86400.;

   dtDest = (double) nDate + ((nDate >= 0) ? dblTime : -dblTime);

   return true;
}

CHMboolean CHMdateTimePrivate::SetTmFromDATE(DATE dtSrc, struct tm& tmDest)
{
   // The legal range does not actually span year 0 to 9999.
   if (dtSrc > MAX_DATE || dtSrc < MIN_DATE) // about year 100 to about 9999
      return false;

   //long nDays;             // Number of days since Dec. 30, 1899
   long nDaysAbsolute;     // Number of days since 1/1/0
   long nSecsInDay;        // Time in seconds since midnight
   long nMinutesInDay;     // Minutes in day

   long n400Years;         // Number of 400 year increments since 1/1/0
   long n400Century;       // Century within 400 year block (0,1,2 or 3)
   long n4Years;           // Number of 4 year increments since 1/1/0
   long n4Day;             // Day within 4 year block
                           //  (0 is 1/1/yr1, 1460 is 12/31/yr4)
   long n4Yr;              // Year within 4 year block (0,1,2 or 3)
   CHMboolean bLeap4 = true;     // true if 4 year block includes leap year

   double dblDate = dtSrc; // tempory serial date

   // If a valid date, then this conversion should not overflow
   //nDays = (long)dblDate;

   // Round to the second
   dblDate += ((dtSrc > 0.0) ? HALF_SECOND : -HALF_SECOND);

   nDaysAbsolute = (long)dblDate + 693959L; // Add days from 1/1/0 to 12/30/1899

   dblDate = fabs(dblDate);
   nSecsInDay = (long)((dblDate - floor(dblDate)) * 86400.);

   // Calculate the day of week (sun=1, mon=2...)
   //   -1 because 1/1/0 is Sat.  +1 because we want 1-based
   tmDest.tm_wday = (int)((nDaysAbsolute - 1) % 7L) + 1;

   // Leap years every 4 yrs except centuries not multiples of 400.
   n400Years = (long)(nDaysAbsolute / 146097L);

   // set nDaysAbsolute to day within 400-year block
   nDaysAbsolute %= 146097L;

   // -1 because first century has extra day
   n400Century = (long)((nDaysAbsolute - 1) / 36524L);

   // Non-leap century
   if (n400Century != 0)
   {
      // set nDaysAbsolute to day within century
      nDaysAbsolute = (nDaysAbsolute - 1) % 36524L;

      // +1 because 1st 4 year increment has 1460 days
      n4Years = (long)((nDaysAbsolute + 1) / 1461L);

      if (n4Years != 0)
         n4Day = (long)((nDaysAbsolute + 1) % 1461L);
      else
      {
         bLeap4 = false;
         n4Day = (long)nDaysAbsolute;
      }
   }
   else
   {
      // Leap century - not special case!
      n4Years = (long)(nDaysAbsolute / 1461L);
      n4Day = (long)(nDaysAbsolute % 1461L);
   }

   if (bLeap4)
   {
      // -1 because first year has 366 days
      n4Yr = (n4Day - 1) / 365;

      if (n4Yr != 0)
         n4Day = (n4Day - 1) % 365;
   }
   else
   {
      n4Yr = n4Day / 365;
      n4Day %= 365;
   }

   // n4Day is now 0-based day of year. Save 1-based day of year, year number
   tmDest.tm_yday = (int)n4Day + 1;
   tmDest.tm_year = n400Years * 400 + n400Century * 100 + n4Years * 4 + n4Yr;

   // Handle leap year: before, on, and after Feb. 29.
   if (n4Yr == 0 && bLeap4)
   {
      // Leap Year
      if (n4Day == 59)
      {
         /* Feb. 29 */
         tmDest.tm_mon = 2;
         tmDest.tm_mday = 29;
         goto DoTime;
      }

      // Pretend it's not a leap year for month/day comp.
      if (n4Day >= 60)
         --n4Day;
   }

   // Make n4DaY a 1-based day of non-leap year and compute
   //  month/day for everything but Feb. 29.
   ++n4Day;

   // Month number always >= n/32, so save some loop time */
   for (tmDest.tm_mon = (n4Day >> 5) + 1;
      n4Day > CHMMonthDays[tmDest.tm_mon]; tmDest.tm_mon++);

   tmDest.tm_mday = (int)(n4Day - CHMMonthDays[tmDest.tm_mon-1]);

DoTime:
   if (nSecsInDay == 0)
      tmDest.tm_hour = tmDest.tm_min = tmDest.tm_sec = 0;
   else
   {
      tmDest.tm_sec = (int)nSecsInDay % 60L;
      nMinutesInDay = nSecsInDay / 60L;
      tmDest.tm_min = (int)nMinutesInDay % 60;
      tmDest.tm_hour = (int)nMinutesInDay / 60;
   }

   return true;
}

CHMdateTime CHMdateTime::Now()
{
   return CHMdateTime(::time(NULL));
}

CHMdateTime::CHMdateTime(time_t Time)
{
   pMember = new CHMdateTimePrivate();
   operator=(Time);
}

CHMdateTime::CHMdateTime(CHMdateTimeHandle Handle)
{
   CHM_ASSERT(Handle != NULL);
   pMember = new CHMdateTimePrivate(Handle);
}

CHMdateTime::CHMdateTime()
{
   pMember = new CHMdateTimePrivate();
}

CHMdateTime::~CHMdateTime()
{
   delete pMember;
}

const CHMdateTimeHandle CHMdateTime::DateTimeHandle() const
{
   return pMember->Handle;
}

CHMdateTime::CHMdateTime(const CHMdateTime& Orig)
{
   pMember = new CHMdateTimePrivate(Orig.pMember->Handle);
}
   
CHMdateTime& CHMdateTime::operator=(const CHMdateTime& Orig)
{
   CHM_CHECK_CALL(CHMdateTimeRelease(pMember->Handle));
   pMember->Handle = Orig.pMember->Handle;
   CHM_CHECK_CALL(CHMdateTimeAddRef(pMember->Handle));
   return *this;
}

CHMdateTime& CHMdateTime::operator=(time_t timeSrc)
{
   // Convert time_t to struct tm
#if defined(WIN32) || defined(_WIN32)
   tm *ptm = localtime(&timeSrc); // ok - windows
#else
   tm TimeStore = {0};
   tm *ptm = localtime_r(&timeSrc,&TimeStore);
#endif

   if (ptm != NULL)
   {
      DATE TheDate;
      pMember->SetDATEFromTm((int)(ptm->tm_year + 1900),
                            (int)(ptm->tm_mon + 1), 
                            (int)ptm->tm_mday,
                            (int)ptm->tm_hour, 
                            (int)ptm->tm_min,
                            (int)ptm->tm_sec, 
                            TheDate);
     SetDate(TheDate);
   }
   else
   {
      throw CHMerror(CHM_TEXT("localtime call failed - timeSrc before 1/1/70 12am"),__LINE__, __FILE__, 0);
  }
  return *this;
} 

DATE CHMdateTime::Date() const
{
   DATE Value;
   CHM_CHECK_CALL(CHMdateTimeGetDATE(pMember->Handle, &Value));
   return Value;
}

void CHMdateTime::SetDate(DATE NewValue)
{
   CHM_CHECK_CALL(CHMdateTimeSetDATE(pMember->Handle, NewValue));
}

double CHMdateTime::SecondFraction() const
{
   double Value;
   CHM_CHECK_CALL(CHMdateTimeGetSecondFraction(pMember->Handle, &Value));
   return Value;
}

void CHMdateTime::SetSecondFraction(double Value)
{
   CHM_CHECK_CALL(CHMdateTimeSetSecondFraction(pMember->Handle, Value));
}

short CHMdateTime::Zone() const
{
   short Value;
   CHM_CHECK_CALL(CHMdateTimeGetZone(pMember->Handle, &Value));
   return Value;
}

void CHMdateTime::SetZone(short Value)
{
   CHM_CHECK_CALL(CHMdateTimeSetZone(pMember->Handle, Value));
}

CHMboolean CHMdateTime::IsNull() const
{
   short Value;
   CHM_CHECK_CALL(CHMdateTimeGetIsNull(pMember->Handle, &Value));
   return Value == 1;
}

void CHMdateTime::setNull()
{
   CHM_CHECK_CALL(CHMdateTimeSetNull(pMember->Handle));
}

CHMboolean CHMdateTime::SetDateTime
(
   int nYear, int nMonth, int nDay,
   int nHour, int nMin,   int nSec
)
{
   DATE TheDate;
   CHMboolean Result = pMember->SetDATEFromTm(nYear, nMonth,
                                nDay, nHour, nMin, nSec, TheDate);
   SetDate(TheDate);
   return Result;
}

CHMboolean CHMdateTime::SetDate
(
   int nYear, 
   int nMonth, 
   int nDay
)
{
  return SetDateTime(nYear, nMonth, nDay, 0, 0, 0);
}

CHMboolean CHMdateTime::SetTime
(   
   int nHour, int nMin, int nSec
)
{
   if (nHour < 0 || nHour > 23 ||
       nMin < 0 || nMin > 59 ||
       nSec < 0 || nSec > 59)
   {
      return false;
   }
       
   SetDate((((long)nHour * 3600L) +  // hrs in seconds
      ((long)nMin * 60L) +  // mins in seconds
      ((long)nSec)) / 86400.);
   return true;
}


int CHMdateTime::Year() const
{
   struct tm tmTemp;

   if (!pMember->SetTmFromDATE(Date(), tmTemp))
   {
      THROW_DATE_TIME_ERROR;
   }
   return tmTemp.tm_year;
}

int CHMdateTime::Month() const 
{
   struct tm tmTemp;

   if (!pMember->SetTmFromDATE(Date(), tmTemp))
   {
      THROW_DATE_TIME_ERROR;
   }
   return tmTemp.tm_mon;
}

int CHMdateTime::Day() const
{
   struct tm tmTemp;

   if (!pMember->SetTmFromDATE(Date(), tmTemp))
   {
      THROW_DATE_TIME_ERROR;
   }
   return tmTemp.tm_mday;
}

int CHMdateTime::Hour() const
{
   struct tm tmTemp;

   if (!pMember->SetTmFromDATE(Date(), tmTemp))
   {
      THROW_DATE_TIME_ERROR;
   }
   
   return tmTemp.tm_hour;
}

int CHMdateTime::Minute() const
{
   struct tm tmTemp;

   if (!pMember->SetTmFromDATE(Date(), tmTemp))
   {
      THROW_DATE_TIME_ERROR;
   }
   return tmTemp.tm_min;
}

int CHMdateTime::Second() const
{
   struct tm tmTemp;

   if (!pMember->SetTmFromDATE(Date(), tmTemp))
   {
      THROW_DATE_TIME_ERROR;
      
   }
   return tmTemp.tm_sec;
}

int CHMdateTime::DayOfWeek() const
{
   struct tm tmTemp;

   if (!pMember->SetTmFromDATE(Date(), tmTemp))
   {
      THROW_DATE_TIME_ERROR;
   }
   return tmTemp.tm_wday;
}

int CHMdateTime::DayOfYear() const
{
   struct tm tmTemp;

   if (!pMember->SetTmFromDATE(Date(), tmTemp))
   {
      THROW_DATE_TIME_ERROR;
   }
   return tmTemp.tm_yday;
}

//void CHMdateTime::PrintOn(ostream& Stream) const
//{
//   Stream << Day() << '/' << Month() << '/' << Year() << ' '
//          << Hour() << ':' << Minute() << ':' << Second() << '.' << "(zone " << Zone() << ')';;
//}

//ostream& operator<<(ostream& Stream, const CHMdateTime& DateTime)
//{
//   DateTime.PrintOn(Stream);
//   return Stream;
//}

CHMstring CHMdateTime::Dump() const
{
   CHMstring Dump(256,'\0');

   sprintf(Dump.get_buffer(),"%d/%d/%d %d:%d:%d. (zone %d)",
           Day(),Month(),Year(),Hour(),Minute(),Second(),Zone());

   return Dump;
}
